from enum import Enum
from pathlib import Path

from yanga.cmake.cmake_backend import CMakePath
from yanga.domain.artifacts import ProjectArtifactsLocator


class BuildArtifact(Enum):
    REPORT_CONFIG = "report_config.json"
    TARGETS_DATA = "targets_data.json"
    COMPILE_COMMANDS = "compile_commands.json"
    COVERAGE_JSON = "coverage.json"
    COVERAGE_DOC = "coverage.md"

    def __init__(self, path: str) -> None:
        self.path = path


class CMakeArtifactsLocator:
    """Defines the paths to the CMake artifacts."""

    def __init__(self, output_dir: Path, project_artifact_locator: ProjectArtifactsLocator) -> None:
        # The directory where the build files will be generated
        self.artifacts_locator = project_artifact_locator
        self.cmake_build_dir = CMakePath(output_dir, "CMAKE_BUILD_DIR")
        self.cmake_project_dir = CMakePath(self.artifacts_locator.project_root_dir)
        self.cmake_variant_reports_dir = self.cmake_build_dir.joinpath("reports")

    @property
    def project_root_dir(self) -> Path:
        return self.artifacts_locator.project_root_dir

    def get_component_build_dir(self, component_name: str) -> CMakePath:
        return self.cmake_build_dir.joinpath(component_name)

    def get_component_reports_dir(self, component_name: str) -> CMakePath:
        return self.get_component_build_dir(component_name).joinpath("reports")

    def get_build_artifact(self, artifact: BuildArtifact) -> CMakePath:
        return self.cmake_build_dir.joinpath(artifact.path)

    def get_component_build_artifact(self, component_name: str, artifact: BuildArtifact) -> CMakePath:
        return self.get_component_build_dir(component_name).joinpath(artifact.path)



from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Optional

from mashumaro import DataClassDictMixin

from yanga.domain.component_analyzer import ComponentAnalyzer
from yanga.domain.execution_context import (
    ExecutionContext,
    UserRequest,
    UserRequestScope,
    UserRequestTarget,
    UserVariantRequest,
)

from .cmake_backend import (
    CMakeAddExecutable,
    CMakeAddLibrary,
    CMakeComment,
    CMakeCustomTarget,
    CMakeElement,
    CMakeIncludeDirectories,
    CMakePath,
    CMakeTargetIncludeDirectories,
    IncludeScope,
)
from .generator import CMakeGenerator


@dataclass
class CreateExecutableConfig(DataClassDictMixin):
    #: If this is enabled, all includes are defined globally and not component specific
    use_global_includes: bool = True


class CreateExecutableCMakeGenerator(CMakeGenerator):
    """Generates CMake elements to build an executable for a variant."""

    def __init__(self, execution_context: ExecutionContext, output_dir: Path, config: Optional[dict[str, Any]] = None) -> None:
        super().__init__(execution_context, output_dir, config)

    @cached_property
    def config_obj(self) -> CreateExecutableConfig:
        """Lazily creates and caches a GTestCMakeGeneratorConfig instance."""
        return CreateExecutableConfig.from_dict(self.config) if self.config else CreateExecutableConfig()

    @property
    def variant_name(self) -> Optional[str]:
        return self.execution_context.variant_name

    def generate(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        elements.append(CMakeComment(f"Generated by {self.__class__.__name__}"))
        elements.extend(self.create_variant_cmake_elements())
        elements.extend(self.create_components_cmake_elements())
        return elements

    def create_variant_cmake_elements(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        if self.config_obj.use_global_includes:
            elements.append(self.get_include_directories())
        else:
            elements.append(CMakeComment("Use global includes for all components disabled."))
        # TODO: I do not like that I have to know here that the components are object libraries
        component_library_targets = [CMakeAddLibrary(component.name).target_name for component in self.execution_context.components]
        variant_executable = CMakeAddExecutable(
            "${PROJECT_NAME}",
            sources=[],
            libraries=component_library_targets,
        )

        elements.append(variant_executable)
        elements.append(
            CMakeCustomTarget(
                name=UserVariantRequest(self.variant_name, UserRequestTarget.BUILD).target_name,
                description=f"Build variant {self.variant_name}",
                commands=[],
                depends=[variant_executable.name],
            )
        )
        elements.append(
            CMakeCustomTarget(
                name=UserVariantRequest(self.variant_name, UserRequestTarget.COMPILE).target_name,
                description=f"Compile variant {self.variant_name}",
                commands=[],
                depends=component_library_targets,
            )
        )

        return elements

    def get_include_directories(self) -> CMakeIncludeDirectories:
        collector = ComponentAnalyzer(
            self.execution_context.components,
            self.execution_context.create_artifacts_locator(),
        )
        include_dirs = collector.collect_include_directories() + self.execution_context.include_directories
        return CMakeIncludeDirectories([CMakePath(path) for path in include_dirs])

    def get_component_include_directories(self, component_analyzer: ComponentAnalyzer) -> list[CMakePath]:
        """Get include directories specific to this component."""
        include_dirs = component_analyzer.collect_include_directories() + self.execution_context.include_directories
        return [CMakePath(path) for path in include_dirs]

    def create_components_cmake_elements(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        for component in self.execution_context.components:
            component_analyzer = ComponentAnalyzer([component], self.execution_context.create_artifacts_locator())
            sources = component_analyzer.collect_sources()
            component_library = CMakeAddLibrary(component.name, sources)
            elements.append(component_library)

            # Add component-specific include directories when global includes are disabled
            if not self.config_obj.use_global_includes:
                include_dirs: list[CMakePath] = self.get_component_include_directories(component_analyzer)
                if include_dirs:
                    # Determine include scope: use PRIVATE for libraries with sources, INTERFACE for header-only
                    scope = IncludeScope.INTERFACE if not sources else IncludeScope.PRIVATE
                    target_includes = CMakeTargetIncludeDirectories(component_library.target_name, include_dirs, scope)
                    elements.append(target_includes)

            elements.append(
                CMakeCustomTarget(
                    UserRequest(
                        UserRequestScope.COMPONENT,
                        self.variant_name,
                        component.name,
                        UserRequestTarget.COMPILE,
                    ).target_name,
                    f"Compile component {component.name}",
                    [],
                    [component_library.target_name],
                )
            )
            elements.append(
                CMakeCustomTarget(
                    UserRequest(
                        UserRequestScope.COMPONENT,
                        self.variant_name,
                        component.name,
                        UserRequestTarget.BUILD,
                    ).target_name,
                    f"Compile component {component.name}",
                    [],
                    [component_library.target_name],
                )
            )
        return elements



"""Utility functions for Sphinx documentation generation which can be used in 'conf.py'."""

import io
import json
import os
import traceback
from pathlib import Path
from typing import Any, Optional

from py_app_dev.core.exceptions import UserNotificationException

from yanga.domain.reports import ReportData


def _relativize_path(file: Path, project_dir: Path) -> str:
    return file.absolute().relative_to(project_dir).as_posix()


class SphinxReportConfig(ReportData):
    @classmethod
    def from_json_file(cls, file_path: Path) -> "SphinxReportConfig":
        try:
            result = cls.from_dict(json.loads(file_path.read_text()))
        except Exception as e:
            output = io.StringIO()
            traceback.print_exc(file=output)
            raise UserNotificationException(output.getvalue()) from e
        return result

    def _relativize_path(self, file: Path) -> str:
        return _relativize_path(file, self.project_dir)

    def create_component_myst_toc(self, component_name: str, maxdepth: int = 1, caption: Optional[str] = None) -> str:
        """Generate the myst table of contents for the component documentation."""
        content = [
            "\n```{toctree}",
            f":maxdepth: {maxdepth}\n",
        ]
        for file in self.get_component_files_list(component_name):
            content.append(f"{file}")
        content.append("```\n")
        return "\n".join(content)

    def _collect_files_from_config(self, config: Any) -> list[str]:
        """Helper method to collect and relativize file paths from a configuration object."""
        content = []
        file_attributes = ["docs_files", "test_results", "coverage_results", "lint_results", "sources", "other_files"]

        for attr in file_attributes:
            if hasattr(config, attr):
                files = getattr(config, attr)
                for file in files:
                    content.append(f"{self._relativize_path(file)}")
        return content

    def get_component_files_list(self, component_name: str) -> list[str]:
        component = next((comp for comp in self.components if comp.name == component_name), None)
        if not component:
            return []
        return self._collect_files_from_config(component)

    def get_variant_files_list(self) -> list[str]:
        if not self.variant_data:
            return []
        return self._collect_files_from_config(self.variant_data)

    @property
    def say_hello(self) -> str:
        return "Hello Sphinx!"


class SphinxConfig:
    """Configuration options for Sphinx documentation generation."""

    REPORT_CONFIGURATION_FILE_ENV_NAME = "REPORT_CONFIGURATION_FILE"

    def __init__(self) -> None:
        self.report_data = self._load_report_config_data()

    def _load_report_config_data(self) -> Optional[SphinxReportConfig]:
        report_config_file_path_str = os.environ.get(self.REPORT_CONFIGURATION_FILE_ENV_NAME, None)

        if report_config_file_path_str:
            report_config_file = Path(report_config_file_path_str)
            if report_config_file.is_file():
                return SphinxReportConfig.from_json_file(report_config_file)
        return None

    @property
    def project(self) -> str:
        return self.report_data.variant_name if self.report_data else "Unknown"

    @property
    def html_context(self) -> dict[str, Any]:
        return {"report_data": self.report_data}

    @property
    def include_patterns(self) -> list[str]:
        """Collect all files from the report config."""
        if self.report_data:
            return [_relativize_path(file, self.report_data.project_dir) for file in self.report_data.collect_all_files()]
        return []

    @staticmethod
    def render_with_jinja(app, docname, source) -> None:  # type: ignore
        """Render pages using jinja templating."""
        # Make sure we're outputting HTML
        if app.builder.format != "html":
            return
        src = source[0]
        rendered = app.builder.templates.render_string(src, app.config.html_context)
        source[0] = rendered
        return



from pathlib import Path
from typing import Optional

from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.find import find_elements_of_type
from py_app_dev.core.logging import logger
from py_app_dev.core.pipeline import PipelineLoader

from yanga.cmake.artifacts_locator import BuildArtifact, CMakeArtifactsLocator
from yanga.domain.execution_context import ExecutionContext
from yanga.domain.reports import ComponentReportData, FeaturesReportRelevantFile, ReportData, ReportRelevantFiles, VariantReportData
from yanga.domain.targets import Target, TargetsData, TargetType

from .cmake_backend import (
    CMakeAddExecutable,
    CMakeAddLibrary,
    CMakeComment,
    CMakeCustomCommand,
    CMakeCustomTarget,
    CMakeMinimumVersion,
    CMakePath,
    CMakeProject,
    CMakeVariable,
)
from .generator import CMakeFile, CMakeGenerator, GeneratedFile, GeneratedFileIf


class CMakeGeneratorReference:
    def __init__(self, group_name: Optional[str], _class: type[CMakeGenerator]) -> None:
        self.group_name = group_name
        self._class = _class

    @property
    def name(self) -> str:
        return self._class.__name__


class CMakeBuildSystemGenerator:
    def __init__(
        self,
        execution_context: ExecutionContext,
        output_dir: Path,
    ):
        self.logger = logger.bind()
        self.execution_context = execution_context
        self.output_dir = output_dir
        # The directory where the CMakeLists.txt file is located
        self.cmake_current_list_dir = CMakePath(self.output_dir, "CMAKE_CURRENT_LIST_DIR")
        self.artifacts_locator = CMakeArtifactsLocator(output_dir, execution_context.create_artifacts_locator())

    @property
    def variant_cmake_file(self) -> CMakePath:
        return self.cmake_current_list_dir.joinpath("variant.cmake")

    @property
    def config_cmake_file(self) -> CMakePath:
        return self.cmake_current_list_dir.joinpath("config.cmake")

    @property
    def report_config_file(self) -> CMakePath:
        return self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG)

    @property
    def variant_name(self) -> str:
        return self.execution_context.variant_name or "MyProject"

    def generate(self) -> list[GeneratedFileIf]:
        files: list[GeneratedFileIf] = []
        cmake_files: list[CMakeFile] = []
        cmake_files.append(self.create_config_cmake_file())
        cmake_files.append(self.create_variant_cmake_file())
        files.extend(cmake_files)
        files.append(self.create_report_config_file())
        files.append(self.create_target_dependencies_file(cmake_files))
        return files

    def create_cmake_lists(self) -> CMakeFile:
        cmake_file = CMakeFile(self.output_dir.joinpath("CMakeLists.txt"))
        cmake_file.append(CMakeMinimumVersion("4.1"))
        platform = self.execution_context.platform
        if platform and platform.toolchain_file:
            cmake_file.append(
                CMakeVariable(
                    "CMAKE_TOOLCHAIN_FILE",
                    CMakePath(self.execution_context.create_artifacts_locator().locate_artifact(platform.toolchain_file, [platform.file])).to_string(),
                )
            )
        cmake_file.append(CMakeProject(self.variant_name))
        return cmake_file

    def create_variant_cmake_file(self) -> CMakeFile:
        cmake_file = CMakeFile(self.variant_cmake_file.to_path())
        cmake_build_dir_var = self.artifacts_locator.cmake_build_dir.to_cmake_element()
        if cmake_build_dir_var:
            cmake_file.append(cmake_build_dir_var)
        # Load all configured CMake generators
        platform = self.execution_context.platform
        if platform:
            try:
                steps_references = PipelineLoader[CMakeGenerator](platform.cmake_generators, self.execution_context.project_root_dir).load_steps()
                for step_reference in steps_references:
                    step = step_reference._class(self.execution_context, self.output_dir, step_reference.config)
                    cmake_file.extend(step.generate())
            except FileNotFoundError as e:
                raise UserNotificationException(e) from e
            except TypeError as e:
                raise UserNotificationException(f"{e}. Please check {platform.file} for {step}.") from e
        return cmake_file

    def create_config_cmake_file(self) -> CMakeFile:
        from .variant_config import ConfigCMakeGenerator

        cmake_file = CMakeFile(self.config_cmake_file.to_path())
        config_generator = ConfigCMakeGenerator(self.execution_context, self.output_dir)
        cmake_file.extend(config_generator.generate())
        cmake_file.append(CMakeComment("Enable generation of compile_commands.json for IDEs and code analysis tools"))
        cmake_file.append(CMakeVariable("CMAKE_EXPORT_COMPILE_COMMANDS", "ON", True, "BOOL", "", True))
        return cmake_file

    def create_report_config_file(self) -> GeneratedFileIf:
        # Collect all ReportRelevantFiles from the data registry
        relevant_files_entries = self.execution_context.data_registry.find_data(ReportRelevantFiles)

        # Group report relevant files by component name. Use a default list
        components_data: dict[str, list[ReportRelevantFiles]] = {}
        variant_data: list[ReportRelevantFiles] = []
        for entry in relevant_files_entries:
            # Only collect data relevant for components
            if entry.target.component_name:
                if components_data.get(entry.target.component_name):
                    components_data[entry.target.component_name].append(entry)
                else:
                    components_data[entry.target.component_name] = [entry]
            else:
                variant_data.append(entry)
        config = ReportData(
            variant_name=self.execution_context.variant_name or "",
            platform_name=self.execution_context.platform.name if self.execution_context.platform else "",
            project_dir=self.execution_context.project_root_dir,
            components=[
                ComponentReportData(
                    name=component_name,
                    files=files,
                    build_dir=self.artifacts_locator.get_component_build_dir(component_name).to_path(),
                )
                for component_name, files in components_data.items()
            ],
        )
        if variant_data:
            config.variant_data = VariantReportData(
                files=variant_data,
                build_dir=self.artifacts_locator.cmake_build_dir.to_path(),
            )
        features_config_file = self.execution_context.data_registry.find_data(FeaturesReportRelevantFile)
        if features_config_file:
            # TODO: Handle multiple feature config files?
            config.features_json_config = features_config_file[0].json_config_file
        return GeneratedFile(self.report_config_file.to_path(), config.to_json_string())

    def create_target_dependencies_file(self, cmake_files: list[CMakeFile]) -> GeneratedFileIf:
        """
        Create a json file that contains the CMake target dependencies tree for the current variant and each component.

        Parse all cmake files, search for the custom targets and their dependencies, and create a json file.
        This file is required to create a graph of the targets and their dependencies.
        """
        targets: list[Target] = []

        # Extract all elements from cmake files
        all_elements = []
        for cmake_file in cmake_files:
            all_elements.extend(cmake_file.content)

        # Find all custom commands
        custom_commands = find_elements_of_type(all_elements, CMakeCustomCommand)
        for custom_command in custom_commands:
            dependencies = []
            outputs = []

            # Extract dependencies
            if custom_command.depends:
                dependencies.extend([str(dep) for dep in custom_command.depends])

            # Extract outputs
            if custom_command.outputs:
                outputs.extend([str(output) for output in custom_command.outputs])

            targets.append(
                Target(
                    name=f"cmd_{custom_command.description.replace(' ', '_').lower()}",
                    description=custom_command.description,
                    depends=dependencies,
                    outputs=outputs,
                    target_type=TargetType.CUSTOM_COMMAND,
                )
            )

        # Find all custom targets
        custom_targets = find_elements_of_type(all_elements, CMakeCustomTarget)
        for custom_target in custom_targets:
            dependencies = []
            outputs = []

            # Extract dependencies
            if custom_target.depends:
                dependencies.extend([str(dep) for dep in custom_target.depends])

            # Extract outputs/byproducts
            if custom_target.byproducts:
                outputs.extend([str(byproduct) for byproduct in custom_target.byproducts])

            targets.append(Target(name=custom_target.name, description=custom_target.description, depends=dependencies, outputs=outputs, target_type=TargetType.CUSTOM_TARGET))

        # Find all executables
        executables = find_elements_of_type(all_elements, CMakeAddExecutable)
        for executable in executables:
            dependencies = []

            # Add library dependencies
            if executable.libraries:
                dependencies.extend(executable.libraries)

            targets.append(
                Target(
                    name=executable.name,
                    description=f"Executable target: {executable.name}",
                    depends=dependencies,
                    outputs=[executable.name],  # The executable itself is the output
                    target_type=TargetType.EXECUTABLE,
                )
            )

        # Find all object libraries
        object_libraries = find_elements_of_type(all_elements, CMakeAddLibrary)
        for obj_lib in object_libraries:
            targets.append(
                Target(
                    name=obj_lib.target_name,
                    description=f"Object library: {obj_lib.name}",
                    depends=[],  # Object libraries typically don't have explicit dependencies in our structure
                    outputs=[obj_lib.target_name],
                    target_type=TargetType.OBJECT_LIBRARY,
                )
            )

        targets_data = TargetsData(targets=targets)
        return GeneratedFile(
            self.cmake_current_list_dir.joinpath("targets_data.json").to_path(),
            targets_data.to_json_string(),
        )



from pathlib import Path
from typing import Any, Optional

from yanga.cmake.artifacts_locator import BuildArtifact, CMakeArtifactsLocator
from yanga.cmake.cmake_backend import CMakeCommand, CMakeComment, CMakeCustomCommand, CMakeCustomTarget, CMakeElement, CMakePath
from yanga.cmake.coverage import CoverageArtifactsLocator
from yanga.cmake.generator import CMakeGenerator
from yanga.docs.sphinx import SphinxConfig
from yanga.domain.component_analyzer import ComponentAnalyzer
from yanga.domain.execution_context import ExecutionContext, UserRequest, UserRequestScope, UserRequestTarget
from yanga.domain.reports import ReportRelevantFiles, ReportRelevantFileType


class ReportCMakeGenerator(CMakeGenerator):
    def __init__(
        self,
        execution_context: ExecutionContext,
        output_dir: Path,
        config: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(execution_context, output_dir, config)
        self.artifacts_locator = CMakeArtifactsLocator(output_dir, execution_context.create_artifacts_locator())

    def generate(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        elements.append(CMakeComment(f"Generated by {self.__class__.__name__}"))
        elements.extend(self.create_components_cmake_elements())
        elements.extend(self.create_variant_cmake_elements())
        return elements

    def create_variant_cmake_elements(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        variant_report_dir = self.artifacts_locator.cmake_variant_reports_dir

        # Create custom command to generate the targets data documentation
        targets_data_doc_file = self.artifacts_locator.cmake_build_dir.joinpath("targets_data.md")
        targets_data_file = self.artifacts_locator.get_build_artifact(BuildArtifact.TARGETS_DATA)
        targets_data_cmd = CMakeCustomCommand(
            description="Generate variant targets data documentation",
            depends=[targets_data_file],
            outputs=[targets_data_doc_file],
            commands=[
                CMakeCommand(
                    "yanga_cmd",
                    [
                        "targets_doc",
                        "--variant-targets-data-file",
                        targets_data_file,
                        "--output-file",
                        targets_data_doc_file,
                    ],
                ),
            ],
        )
        elements.append(targets_data_cmd)

        # Register the variant targets data documentation file as relevant for the variant report
        variant_user_request = UserRequest(UserRequestScope.VARIANT, target=UserRequestTarget.DOCS)
        self.execution_context.data_registry.insert(
            ReportRelevantFiles(
                target=variant_user_request,
                files_to_be_included=[targets_data_doc_file.to_path()],
                file_type=ReportRelevantFileType.OTHER,
            ),
            variant_user_request.target_name,
        )
        # Check if there are any coverage reports registered for the variant
        coverage_reports = any(
            entry
            for entry in self.execution_context.data_registry.find_data(ReportRelevantFiles)
            if entry.file_type == ReportRelevantFileType.COVERAGE_RESULT and entry.target.scope == UserRequestScope.VARIANT
        )

        # Create variant results target to collect all component and variant results relevant for the report
        results_target_depends: list[str | CMakePath] = [
            self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG),
            *[
                UserRequest(UserRequestScope.COMPONENT, target=UserRequestTarget.RESULTS, component_name=component.name).target_name
                for component in self.execution_context.components
            ],
        ]
        if targets_data_cmd.outputs:
            results_target_depends.extend(targets_data_cmd.outputs)
        if coverage_reports:
            results_target_depends.append(UserRequest(UserRequestScope.VARIANT, target=UserRequestTarget.COVERAGE).target_name)
            # The html coverage reports are generated in the component specific reports directories.
            # We need to create custom commands to copy them to the variant report directory.
            components_with_coverage_results = [
                entry.target.component_name
                for entry in self.execution_context.data_registry.find_data(ReportRelevantFiles)
                if entry.file_type == ReportRelevantFileType.COVERAGE_RESULT and entry.target.scope == UserRequestScope.COMPONENT and entry.target.component_name is not None
            ]
            for component_name in components_with_coverage_results:
                artifacts_locator = CoverageArtifactsLocator.from_cmake_artifacts_locator(self.artifacts_locator)
                # Get the directory where the component coverage html report was generated. This directory will be copied to the variant report directory
                component_coverage_html_dir = artifacts_locator.get_component_coverage_html_dir(component_name)
                coverage_doc_file = artifacts_locator.get_component_build_artifact(component_name, BuildArtifact.COVERAGE_DOC)
                # We need to find where will the coverage doc file html result will be generates in the variant report directory
                # The component coverage doc file is in the build directory and will be included by the sphinx build with its relative path to the project root
                # This means that the coverage.html generated file will have the relative path to the project root too inside the report directory
                # Now we need to determine the coverage.html path based on the coverage.md path and the sphinx output directory
                coverage_doc_file_relative_path = coverage_doc_file.to_path().relative_to(self.artifacts_locator.project_root_dir).parent.as_posix()
                component_variant_coverage_html_dir = variant_report_dir.joinpath(coverage_doc_file_relative_path).joinpath("coverage")
                # Create custom command to copy the component coverage html report to the variant report directory
                copy_coverage_html_cmd = CMakeCustomCommand(
                    description=f"Copy coverage html report for component {component_name} to variant report directory",
                    depends=[],
                    outputs=[component_variant_coverage_html_dir],
                    commands=[
                        CMakeCommand(
                            "${CMAKE_COMMAND}",
                            [
                                "-E",
                                "copy_directory",
                                component_coverage_html_dir,
                                component_variant_coverage_html_dir,
                            ],
                        ),
                    ],
                )
                elements.append(copy_coverage_html_cmd)
                if copy_coverage_html_cmd.outputs:
                    results_target_depends.extend(copy_coverage_html_cmd.outputs)

        results_target = CMakeCustomTarget(
            name=UserRequest(
                UserRequestScope.VARIANT,
                target=UserRequestTarget.RESULTS,
            ).target_name,
            description=f"Run all targets for all component results for {self.execution_context.variant_name}",
            commands=[],
            depends=results_target_depends,
        )
        elements.append(results_target)
        elements.append(
            CMakeCustomTarget(
                name=UserRequest(
                    UserRequestScope.VARIANT,
                    target=UserRequestTarget.REPORT,
                ).target_name,
                description=f"Run sphinx build for variant {self.execution_context.variant_name}",
                commands=[
                    CMakeCommand(
                        "${CMAKE_COMMAND}",
                        [
                            "-E",
                            "make_directory",
                            variant_report_dir,
                        ],
                    ),
                    CMakeCommand(
                        "${CMAKE_COMMAND}",
                        [
                            "-E",
                            "env",
                            f"{SphinxConfig.REPORT_CONFIGURATION_FILE_ENV_NAME}={self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG)}",
                            "--",
                            "sphinx-build",
                            "-E",
                            "-b",
                            "html",
                            self.artifacts_locator.cmake_project_dir,
                            variant_report_dir,
                        ],
                    ),
                ],
                depends=[
                    self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG),
                    results_target.name,
                ],
            )
        )
        return elements

    def create_components_cmake_elements(self) -> list[CMakeElement]:
        elements: list[CMakeElement] = []
        for component in self.execution_context.components:
            component_analyzer = ComponentAnalyzer([component], self.execution_context.create_artifacts_locator())
            component_build_dir = self.artifacts_locator.get_component_build_dir(component.name)
            report_config_output_file = self.artifacts_locator.get_component_build_artifact(component.name, BuildArtifact.REPORT_CONFIG)
            source_files: list[CMakePath] = []
            if not (component.docs and component.docs.exclude_productive_code):
                source_files.extend([CMakePath(source) for source in component_analyzer.collect_sources()])

            # Check if there are any test results registered for the component
            test_results = any(
                entry
                for entry in self.execution_context.data_registry.find_data(ReportRelevantFiles)
                # TODO: when test results are registered, use the TEST_RESULT type.
                #       Currently, we assume that tests were executed if there are coverage results
                if entry.file_type == ReportRelevantFileType.COVERAGE_RESULT and entry.target.component_name == component.name
            )
            if test_results:
                source_files.extend([CMakePath(source) for source in component_analyzer.collect_test_sources()])
            source_files_output_md = [component_build_dir.joinpath(f"{source_file.to_path().name}.md") for source_file in source_files]
            component_docs_target = UserRequest(
                UserRequestScope.COMPONENT,
                target=UserRequestTarget.DOCS,
                component_name=component.name,
            )

            if source_files_output_md:
                elements.append(
                    CMakeCustomTarget(
                        name=component_docs_target.target_name,
                        description=f"Generate sources docs for component {component.name}",
                        commands=[
                            CMakeCommand(
                                "clanguru",
                                [
                                    "docs",
                                    "--source-file",
                                    source_file,
                                    "--output-file",
                                    output_md,
                                    "--compilation-database",
                                    self.artifacts_locator.get_build_artifact(BuildArtifact.COMPILE_COMMANDS),
                                    "--format",
                                    "myst",
                                ],
                            )
                            for source_file, output_md in zip(source_files, source_files_output_md)
                        ],
                        depends=[self.artifacts_locator.get_build_artifact(BuildArtifact.COMPILE_COMMANDS)],
                        byproducts=source_files_output_md,
                    )
                )
                # Register the component sources md files as relevant for the component report
                self.execution_context.data_registry.insert(
                    ReportRelevantFiles(
                        target=component_docs_target,
                        files_to_be_included=[md_output.to_path() for md_output in source_files_output_md],
                        file_type=ReportRelevantFileType.SOURCES,
                    ),
                    component_docs_target.target_name,
                )

            docs_source_files = [CMakePath(source) for source in component_analyzer.collect_docs_sources()]
            if docs_source_files:
                # Register the component documentation files as relevant for the component report
                self.execution_context.data_registry.insert(
                    ReportRelevantFiles(
                        target=component_docs_target,
                        files_to_be_included=[source.to_path() for source in docs_source_files],
                        file_type=ReportRelevantFileType.DOCS,
                    ),
                    component_docs_target.target_name,
                )

            component_results_target = UserRequest(
                UserRequestScope.COMPONENT,
                target=UserRequestTarget.RESULTS,
                component_name=component.name,
            )

            # Collect all component targets that registered files relevant for the report
            result_targets = [entry.target for entry in self.execution_context.data_registry.find_data(ReportRelevantFiles) if entry.target.component_name == component.name]
            # Make list unique and keep order
            result_targets = list(dict.fromkeys(result_targets))

            elements.append(
                CMakeCustomTarget(
                    name=component_results_target.target_name,
                    description=f"Execute targets to get all results for component {component.name}",
                    commands=[],
                    depends=[
                        self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG),
                        *[target.target_name for target in result_targets],
                    ],
                )
            )
            component_report_target = UserRequest(
                UserRequestScope.COMPONENT,
                target=UserRequestTarget.REPORT,
                component_name=component.name,
            )

            component_report_dir = self.artifacts_locator.get_component_reports_dir(component.name)
            elements.append(
                CMakeCustomTarget(
                    name=component_report_target.target_name,
                    description=f"Generate report for component {component.name}",
                    commands=[
                        CMakeCommand(
                            "${CMAKE_COMMAND}",
                            [
                                "-E",
                                "make_directory",
                                component_report_dir,
                            ],
                        ),
                        CMakeCommand(
                            "yanga_cmd",
                            [
                                "report_config",
                                "--component-name",
                                component.name,
                                "--output-file",
                                report_config_output_file,
                                "--variant-report-config",
                                self.artifacts_locator.get_build_artifact(BuildArtifact.REPORT_CONFIG),
                            ],
                        ),
                        CMakeCommand(
                            "${CMAKE_COMMAND}",
                            [
                                "-E",
                                "env",
                                f"REPORT_CONFIGURATION_FILE={report_config_output_file}",
                                "--",
                                "sphinx-build",
                                "-E",
                                "-b",
                                "html",
                                self.artifacts_locator.cmake_project_dir,
                                component_report_dir,
                            ],
                        ),
                    ],
                    depends=[
                        component_results_target.target_name,
                    ],
                )
            )

        return elements



from pathlib import Path

from kspl.config_slurper import KConfigData, VariantData, VariantViewData
from kspl.gui import KSPL
from kspl.kconfig import EditableConfigElement, KConfig
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger
from py_app_dev.mvp.event_manager import EventManager

from yanga.domain.project_slurper import YangaProjectSlurper
from yanga.gui.icons import Icons


class YangaKConfigData(KConfigData):
    """Yanga implementation of KConfigData protocol for viewing feature configurations."""

    def __init__(self, project_dir: Path) -> None:
        self.project_dir = project_dir.absolute()
        self.project_slurper = YangaProjectSlurper(project_dir)
        self.logger = logger.bind()

        # Load KConfig model if it exists
        self.kconfig_model_file = self.project_dir / "KConfig"
        if not self.kconfig_model_file.is_file():
            raise UserNotificationException(f"KConfig file not found at {self.kconfig_model_file}")

        self.base_kconfig = KConfig(self.kconfig_model_file)
        self.variant_kconfigs: list[VariantData] = []
        self._load_variant_configs()

    def _load_variant_configs(self) -> None:
        """Load KConfig configurations for all variants."""
        self.variant_kconfigs = []

        for variant_config in self.project_slurper.variants:
            if variant_config.features_selection_file:
                # Get the features selection file path
                features_file = self.project_slurper.get_variant_config_file(variant_config.name)
                if features_file and features_file.exists():
                    kconfig = KConfig(self.kconfig_model_file, features_file)
                    self.variant_kconfigs.append(VariantData(variant_config.name, kconfig))
                else:
                    self.logger.warning(f"Features selection file not found for variant {variant_config.name}: {variant_config.features_selection_file}")
                    # Use base configuration without selection file
                    kconfig = KConfig(self.kconfig_model_file)
                    self.variant_kconfigs.append(VariantData(variant_config.name, kconfig))
            else:
                # Use base configuration for variants without features selection
                kconfig = KConfig(self.kconfig_model_file)
                self.variant_kconfigs.append(VariantData(variant_config.name, kconfig))

        # If no variants found, create a default one
        if not self.variant_kconfigs:
            self.variant_kconfigs.append(VariantData("Default", self.base_kconfig))

    def get_elements(self) -> list[EditableConfigElement]:
        """Get all configuration elements from the base KConfig model."""
        return self.base_kconfig.elements

    def get_variants(self) -> list[VariantViewData]:
        """Get variant view data for all loaded variants."""
        variants = []
        for variant in self.variant_kconfigs:
            config_dict = {config_elem.name: config_elem.value for config_elem in variant.config.elements if not config_elem.is_menu}
            variants.append(VariantViewData(variant.name, config_dict))
        return variants

    def find_variant_config(self, variant_name: str) -> VariantData | None:
        """Find variant configuration by name."""
        for variant in self.variant_kconfigs:
            if variant.name == variant_name:
                return variant
        return None

    def refresh_data(self) -> None:
        """Refresh the configuration data by reloading everything."""
        self.logger.info("Refreshing Yanga KConfig data...")

        # Reload project configuration
        self.project_slurper = YangaProjectSlurper(self.project_dir)

        # Reload KConfig model
        self.base_kconfig = KConfig(self.kconfig_model_file)

        # Reload variant configurations
        self._load_variant_configs()

        self.logger.info(f"Refreshed data: found {len(self.variant_kconfigs)} variants")


class KConfigView:
    """View for displaying the KConfig data for all variants."""

    def __init__(self, project_dir: Path) -> None:
        self.project_dir = project_dir
        self.logger = logger.bind()

    def run(self) -> None:
        """Run the KConfig viewer GUI."""
        try:
            event_manager = EventManager()
            kconfig_data: KConfigData = YangaKConfigData(self.project_dir)
            KSPL(event_manager, kconfig_data, icon_file=Icons.YANGA_ICON.file).run()
        except UserNotificationException as e:
            self.logger.error(f"Failed to start KConfig viewer: {e}")
        except Exception as e:
            self.logger.error(f"Unexpected error starting KConfig viewer: {e}")
            raise



"""
Command line utility to get create a component specific compile_commands.json.

It gets from the command line:
- path to compile_commands.json
- source-file which can be use multiple times
- output path
"""

import os
import textwrap
from argparse import ArgumentParser, Namespace
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import auto
from pathlib import Path
from typing import Union

from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger

from yanga.cmake.artifacts_locator import BuildArtifact
from yanga.cmake.generator import GeneratedFile
from yanga.domain.config import BaseConfigJSONMixin, StringableEnum
from yanga.domain.reports import ReportData

from .base import create_config


class GcovReportScope(StringableEnum):
    COMPONENT = auto()
    VARIANT = auto()


def _deserialize_component_objects(raw: object) -> list[Path]:
    """
    Normalize various incoming representations (CMake / argparse / mashumaro) into list[Path].

    Accepted forms:
      - "a.obj;b.obj;c.obj"
      - "a.obj"
      - Path("a.obj")
      - [ "a.obj", "b.obj" ]
      - [ Path("a.obj"), Path("b.obj") ]
      - Nested lists produced by earlier parsing: [[Path("a.obj"), ...], ...]
    """

    def iter_items(value: object) -> Iterator[Union[str, Path]]:
        if isinstance(value, str):
            # CMake style list or single path
            yield from (p for p in value.split(";") if p)
        elif isinstance(value, Path):
            yield value
        elif isinstance(value, list):
            for entry in value:
                yield from iter_items(entry)
        else:
            raise TypeError(f"Unsupported component_objects entry type: {type(value).__name__}")

    return [Path(item) for item in iter_items(raw)]


@dataclass
class ComponentCommandArgs(BaseConfigJSONMixin):
    component_objects: list[Path] = field(
        metadata={
            "help": "List of object files for the component",
            "deserialize": _deserialize_component_objects,
        }
    )
    source_files: list[Path] = field(
        metadata={"help": "Relevant source files."},
    )
    output_file: Path = field(
        metadata={"help": "Output Gcovr configuration file."},
    )


class CreateComponentGcovrConfigCommand(Command):
    def __init__(self) -> None:
        super().__init__("gcovr_config_component", "Create a component specific gcovr configuration file.")
        self.logger = logger.bind()

    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = create_config(ComponentCommandArgs, args)
        # Determine the object directory as the common parent of the object files
        if config.component_objects:
            object_directory = os.path.commonpath([obj.parent for obj in config.component_objects])
        else:
            self.logger.error("No component object files provided.")
            return 1
        # Create a gcovr config file
        gcovr_cfg_lines = [
            f"root = {Path(object_directory).as_posix()}",
            *[f"filter = {source_file.as_posix()}" for source_file in config.source_files],
        ]

        GeneratedFile(config.output_file, "\n".join(gcovr_cfg_lines) + "\n").to_file()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, ComponentCommandArgs)


@dataclass
class VariantCommandArgs(BaseConfigJSONMixin):
    variant_report_config: Path = field(metadata={"help": "Variant report configuration"})
    output_file: Path = field(metadata={"help": "Output Gcovr configuration file."})


class CreateVariantGcovrConfigCommand(Command):
    def __init__(self) -> None:
        super().__init__("gcovr_config_variant", "Create a variant gcovr configuration file to collect all components json reports.")
        self.logger = logger.bind()

    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = create_config(VariantCommandArgs, args)

        report_config = ReportData.from_json_file(config.variant_report_config)

        # Only include components which have coverage results
        coverage_json_files = [component.build_dir.joinpath(BuildArtifact.COVERAGE_JSON.path) for component in report_config.components if component.coverage_results]

        # Create a gcovr config file
        gcovr_cfg_lines = [
            f"root = {report_config.project_dir.as_posix()}",
            *[f"add-tracefile = {coverage_json.as_posix()}" for coverage_json in coverage_json_files],
        ]

        GeneratedFile(config.output_file, "\n".join(gcovr_cfg_lines) + "\n").to_file()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, VariantCommandArgs)


@dataclass
class DocCommandArgs(BaseConfigJSONMixin):
    output_file: Path = field(
        metadata={"help": "Output doc file."},
    )


class GcovrDocCommand(Command):
    def __init__(self) -> None:
        super().__init__("gcovr_doc", "Create a component documentation file to include the gcovr generated report.")
        self.logger = logger.bind()

    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = create_config(DocCommandArgs, args)
        GeneratedFile(
            config.output_file,
            textwrap.dedent("""\
                        # Code Coverage

                        <a href="./coverage/index.html">Report</a>
                        """),
        ).to_file()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, DocCommandArgs)



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Optional, Union

from py_app_dev.core.subprocess import SubprocessExecutor
from pypeline.domain.execution_context import ExecutionContext as _ExecutionContext

from .artifacts import ProjectArtifactsLocator
from .components import Component
from .config import PlatformConfig, VariantConfig


class UserRequestTarget(Enum):
    NONE = auto()
    ALL = auto()
    BUILD = auto()
    COMPILE = auto()
    CLEAN = auto()
    TEST = auto()
    COVERAGE = auto()
    MOCKUP = auto()
    LINT = auto()
    DOCS = auto()
    REPORT = auto()
    #: These are all results relevant for reports generation (e.g., test results, coverage results, lint results, etc.)
    RESULTS = auto()

    def __str__(self) -> str:
        return self.name.lower()


class UserRequestScope(Enum):
    VARIANT = auto()
    COMPONENT = auto()


@dataclass(frozen=True, order=True)
class UserRequest:
    scope: UserRequestScope
    variant_name: Optional[str] = None
    component_name: Optional[str] = None
    target: Optional[Union[str, UserRequestTarget]] = None
    build_type: Optional[str] = None

    @property
    def target_name(self) -> str:
        target = str(self.target if self.target else UserRequestTarget.ALL)
        if self.component_name:
            return f"{self.component_name}_{target}"
        return target


class UserVariantRequest(UserRequest):
    def __init__(
        self,
        variant_name: Optional[str],
        target: Optional[Union[str, UserRequestTarget]] = None,
        build_type: Optional[str] = None,
    ) -> None:
        super().__init__(UserRequestScope.VARIANT, variant_name, None, target=target, build_type=build_type)


class IncludeDirectoriesProvider(ABC):
    @abstractmethod
    def get_include_directories(self) -> list[Path]: ...


@dataclass
class ExecutionContext(_ExecutionContext):
    def __init__(
        self,
        project_root_dir: Path,
        user_request: UserRequest,
        variant_name: Optional[str] = None,
        components: Optional[list[Component]] = None,
        user_config_files: Optional[list[Path]] = None,
        features_selection_file: Optional[Path] = None,
        platform: Optional[PlatformConfig] = None,
        variant: Optional[VariantConfig] = None,
        create_yanga_build_dir: bool = True,
    ) -> None:
        super().__init__(project_root_dir)
        self.user_request = user_request
        self.variant_name = variant_name
        self.components = components if components else []
        self.user_config_files = user_config_files if user_config_files else []
        self.features_selection_file = features_selection_file
        self.platform = platform
        self.variant = variant
        self.create_yanga_build_dir = create_yanga_build_dir
        self.include_dirs_providers: list[IncludeDirectoriesProvider] = []

    @property
    def include_directories(self) -> list[Path]:
        include_dirs = []
        for provider in self.include_dirs_providers:
            include_dirs.extend(provider.get_include_directories())
        return include_dirs

    def add_install_dirs(self, install_dirs: list[Path]) -> None:
        self.install_dirs.extend(install_dirs)

    def add_include_dirs_provider(self, provider: IncludeDirectoriesProvider) -> None:
        self.include_dirs_providers.append(provider)

    def create_process_executor(self, command: list[str | Path], cwd: Optional[Path] = None) -> SubprocessExecutor:
        env = os.environ.copy()
        # Update the environment variables with the ones from the execution context
        env.update(self.env_vars)
        # Add the install directories to the PATH
        env["PATH"] = os.pathsep.join([path.absolute().as_posix() for path in self.install_dirs] + [env["PATH"]])
        return SubprocessExecutor(command, cwd=cwd, env=env, shell=True)  # noqa: S604

    def create_artifacts_locator(self) -> ProjectArtifactsLocator:
        return ProjectArtifactsLocator(
            self.project_root_dir,
            self.variant_name,
            self.platform.name if self.platform else None,
            self.user_request.build_type,
            create_yanga_build_dir=self.create_yanga_build_dir,
        )



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional

from yanga.domain.execution_context import ExecutionContext

from .cmake_backend import CMakeElement


class CMakeGenerator(ABC):
    """Base class for CMake generators."""

    def __init__(self, execution_context: ExecutionContext, output_dir: Path, config: Optional[dict[str, Any]] = None) -> None:
        self.execution_context = execution_context
        self.config = config
        self.output_dir = output_dir

    @abstractmethod
    def generate(self) -> list[CMakeElement]:
        pass


class GeneratedFileIf(ABC):
    def __init__(self, path: Path) -> None:
        self.path = path

    @abstractmethod
    def to_string(self) -> str:
        pass

    def to_file(self) -> None:
        dir = self.path.parent
        if not dir.exists():
            dir.mkdir(parents=True, exist_ok=True)
        with open(self.path, "w") as f:
            f.write(self.to_string())


class GeneratedFile(GeneratedFileIf):
    def __init__(self, path: Path, content: str) -> None:
        super().__init__(path)
        self.content = content

    def to_string(self) -> str:
        return self.content


class CMakeFile(GeneratedFileIf):
    def __init__(self, path: Path) -> None:
        super().__init__(path)
        self.content: list[CMakeElement] = []

    def to_string(self) -> str:
        return "\n".join(str(elem) for elem in self.content)

    def append(self, content: Optional[CMakeElement]) -> "CMakeFile":
        if content:
            self.content.append(content)
        return self

    def extend(self, content: list[CMakeElement]) -> "CMakeFile":
        for element in content:
            self.append(element)
        return self



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Any, Optional, Union


def make_list_unique(seq: list[Any]) -> list[Any]:
    return list(dict.fromkeys(seq))


class LibraryType(Enum):
    OBJECT = auto()
    INTERFACE = auto()


class IncludeScope(Enum):
    INTERFACE = auto()
    PUBLIC = auto()
    PRIVATE = auto()


class CMakeElement(ABC):
    tab_prefix = " " * 4

    @abstractmethod
    def to_string(self) -> str:
        pass

    def __str__(self) -> str:
        return self.to_string()


class CMakeContent(CMakeElement):
    def __init__(self, content: str) -> None:
        super().__init__()
        self.content = content

    def to_string(self) -> str:
        return self.content


class CMakeEmptyLine(CMakeElement):
    def to_string(self) -> str:
        return ""


class CMakeComment(CMakeElement):
    def __init__(self, comment: str) -> None:
        super().__init__()
        self.comment = comment

    def to_string(self) -> str:
        return f"# {self.comment}"


class CMakeProject(CMakeElement):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def to_string(self) -> str:
        return f"project({self.name})"


class CMakeMinimumVersion(CMakeElement):
    def __init__(self, version: str) -> None:
        super().__init__()
        self.version = version

    def to_string(self) -> str:
        return f"cmake_minimum_required(VERSION {self.version})"


class CMakeAddLibrary(CMakeElement):
    def __init__(
        self,
        name: str,
        files: list[Path] | None = None,
        type: LibraryType | None = None,
        compile_options: list[str] | None = None,
    ) -> None:
        self.name = name
        self.files = files or []
        self.type = type or (LibraryType.OBJECT if self.files else LibraryType.INTERFACE)
        self.compile_options = compile_options or []

    @property
    def target_name(self) -> str:
        return f"{self.name}_lib"

    def to_string(self) -> str:
        content = f"add_library({self.target_name} {self.type.name} {self._get_files_string()})"
        if self.compile_options:
            content += "\n" + self._add_compile_options()
        return content

    def _get_files_string(self) -> str:
        return " ".join([file.as_posix() for file in self.files])

    def _add_compile_options(self) -> str:
        return f"target_compile_options({self.target_name} PRIVATE " + " ".join(self.compile_options) + ")"


@dataclass
class CMakeVariable(CMakeElement):
    name: str
    value: str
    cache: bool = False
    type: Optional[str] = None
    docstring: Optional[str] = None
    force: bool = False

    def to_string(self) -> str:
        arguments = [self.name, self.value]
        if self.cache:
            arguments.append("CACHE")
        if self.type:
            arguments.append(self.type)
        if self.docstring is not None:
            if self.docstring:
                arguments.append(self.docstring)
            else:
                arguments.append('""')
        if self.force:
            arguments.append("FORCE")
        return "set(" + " ".join(arguments) + ")"


class CMakePath:
    def __init__(
        self,
        path: Path,
        variable: Optional[str] = None,
        relative_path: Optional[Path] = None,
    ) -> None:
        super().__init__()
        self.path = path
        self.variable = variable
        self.relative_path = relative_path

    def to_cmake_element(self) -> Optional[CMakeElement]:
        return CMakeVariable(self.variable, self.to_path().as_posix()) if self.variable else None

    def to_string(self) -> str:
        if self.variable:
            path = f"${{{self.variable}}}"
        else:
            path = self.path.as_posix()
        if self.relative_path:
            path += f"/{self.relative_path.as_posix()}"
        return path

    def to_path(self) -> Path:
        return self.path / self.relative_path if self.relative_path else self.path

    def joinpath(self, path: str) -> "CMakePath":
        rel_path = self.relative_path / path if self.relative_path else Path(path)
        return CMakePath(self.path, self.variable, rel_path)

    def __str__(self) -> str:
        return self.to_string()

    def with_suffix(self, suffix: str) -> "CMakePath":
        if self.relative_path:
            new_relative_path = self.relative_path.with_suffix(suffix)
            return CMakePath(self.path, self.variable, new_relative_path)
        else:
            new_path = self.path.with_suffix(suffix)
            return CMakePath(new_path, self.variable, None)


class CMakeInclude(CMakeElement):
    def __init__(self, path: str | CMakePath) -> None:
        super().__init__()
        self.path = path

    def to_string(self) -> str:
        return f"include({self.path})"


class CMakeIncludeDirectories(CMakeElement):
    def __init__(self, paths: list[CMakePath]) -> None:
        super().__init__()
        self.paths = paths

    def to_string(self) -> str:
        return "\n".join(["include_directories(", *self._add_tabulated_paths(self.paths), ")"])

    def _add_tabulated_paths(self, paths: list[CMakePath]) -> list[str]:
        return [self._add_tabulated_path(path) for path in paths]

    def _add_tabulated_path(self, path: CMakePath) -> str:
        return self.tab_prefix + path.to_string()


class CMakeTargetIncludeDirectories(CMakeElement):
    def __init__(self, target_name: str, paths: list[CMakePath], scope: IncludeScope = IncludeScope.PRIVATE) -> None:
        super().__init__()
        self.target_name = target_name
        self.paths = paths
        self.scope = scope

    def to_string(self) -> str:
        if not self.paths:
            return ""
        paths_str = " ".join(path.to_string() for path in self.paths)
        return f"target_include_directories({self.target_name} {self.scope.name} {paths_str})"


@dataclass
class CMakeAddExecutable(CMakeElement):
    name: str
    sources: list[Union[str, CMakePath, CMakeAddLibrary]]
    libraries: list[str] = field(default_factory=list)
    compile_options: list[str] = field(default_factory=list)
    link_options: list[str] = field(default_factory=list)
    exclude_from_all: bool = False

    def to_string(self) -> str:
        content = self._add_executable()
        if self.libraries:
            # add target_link_libraries
            content += "\n" + self._add_target_link_libraries()
        if self.compile_options:
            content += "\n" + self._add_compile_options()
        if self.link_options:
            content += "\n" + self._add_link_options()
        return content

    def _add_executable(self) -> str:
        arguments = [self.name]
        if self.exclude_from_all:
            arguments.append("EXCLUDE_FROM_ALL")
        arguments.extend(self._get_sources())
        return "add_executable(" + " ".join(arguments) + ")"

    def _add_target_link_libraries(self) -> str:
        return "target_link_libraries(" + " ".join([self.name, *self.libraries]) + ")"

    def _get_sources(self) -> list[str]:
        return [self._get_source(source) for source in self.sources]

    def _get_source(self, source: str | CMakePath | CMakeAddLibrary) -> str:
        """
        Get the source as string.

        - str: return as is
        - CMakePath: return the path as string
        - CMakeAddLibrary: return the target objects
        """
        if isinstance(source, CMakeAddLibrary):
            return f"$<TARGET_OBJECTS:{source.target_name}>"
        else:
            return str(source)

    def _add_compile_options(self) -> str:
        return f"target_compile_options({self.name} PRIVATE " + " ".join(self.compile_options) + ")"

    def _add_link_options(self) -> str:
        return f"target_link_options({self.name} PRIVATE " + " ".join(self.link_options) + ")"


@dataclass
class CMakeSetTargetProperties(CMakeElement):
    target: str
    properties: dict[str, str | CMakePath]

    def to_string(self) -> str:
        if not self.properties:
            return ""

        props = []
        for key, value in self.properties.items():
            props.extend([key, str(value)])

        return f"set_target_properties({self.target} PROPERTIES {' '.join(props)})"


class CMakeCommand(CMakeElement):
    def __init__(self, command: str | CMakePath, arguments: list[str | CMakePath]) -> None:
        super().__init__()
        self.command = command
        self.arguments = arguments

    def to_string(self) -> str:
        result = f"COMMAND {self.command}"
        if self.arguments:
            args_str = " ".join(str(arg) for arg in self.arguments)
            result += f" {args_str}"
        return result


class CMakeExecuteProcess(CMakeElement):
    def __init__(self, description: str, commands: list[CMakeCommand]) -> None:
        super().__init__()
        self.description = description
        self.commands = commands

    def to_string(self) -> str:
        content = [CMakeComment(self.description), "execute_process("]
        content.extend(self._get_commands())
        content.extend(
            [
                f"{self.tab_prefix}RESULT_VARIABLE result",
                ")",
                "if(result)",
                f"{self.tab_prefix}message(FATAL_ERROR '{self.description} failed: ${{result}}')",
                "endif()",
            ]
        )
        return "\n".join(str(line) for line in content)

    def _get_commands(self) -> list[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


class CMakeDepends(CMakeElement):
    def __init__(self, depends: Sequence[str | CMakePath]) -> None:
        super().__init__()
        self.depends = depends

    def to_string(self) -> str:
        return f"{self.tab_prefix}DEPENDS {' '.join(str(depend) for depend in self.depends)}"


class CMakeByproducts(CMakeElement):
    def __init__(self, byproducts: list[CMakePath]) -> None:
        super().__init__()
        self.byproducts = byproducts

    def to_string(self) -> str:
        return f"{self.tab_prefix}BYPRODUCTS {' '.join(str(byproduct) for byproduct in self.byproducts)}"


class CMakeBuildEvent(Enum):
    PRE_BUILD = auto()
    PRE_LINK = auto()
    POST_BUILD = auto()


@dataclass
class CMakeCustomCommand(CMakeElement):
    description: str
    commands: list[CMakeCommand]
    outputs: Optional[list[CMakePath]] = None
    depends: Optional[Sequence[str | CMakePath]] = None
    working_directory: Optional[CMakePath] = None
    build_event: Optional[CMakeBuildEvent] = None
    byproducts: Optional[list[CMakePath]] = None
    target: Optional[str] = None
    command_expand_lists: bool = False

    def to_string(self) -> str:
        content = [CMakeComment(self.description), "add_custom_command("]
        if self.target:
            content.append(f"{self.tab_prefix}TARGET {self.target}")
        if self.build_event:
            content.append(f"{self.tab_prefix}{self.build_event.name}")
        if self.outputs:
            content.extend(self._get_outputs())
        if self.depends:
            content.append(CMakeDepends(self.depends).to_string())
        content.extend(self._get_commands())
        if self.byproducts:
            content.append(CMakeByproducts(self.byproducts).to_string())
        if self.working_directory:
            content.append(f"{self.tab_prefix}WORKING_DIRECTORY {self.working_directory.to_string()}")
        if self.command_expand_lists:
            content.append(f"{self.tab_prefix}COMMAND_EXPAND_LISTS")

        content.append(")")
        return "\n".join(str(line) for line in content)

    def _get_outputs(self) -> list[str]:
        if not self.outputs:
            return []
        return [f"{self.tab_prefix}OUTPUT {' '.join([output.to_string() for output in self.outputs])}"]

    def _get_commands(self) -> list[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


@dataclass
class CMakeCustomTarget(CMakeElement):
    def __init__(
        self,
        name: str,
        description: str,
        commands: list[CMakeCommand],
        depends: Optional[Sequence[str | CMakePath]] = None,
        default_target: bool = False,
        byproducts: Optional[list[CMakePath]] = None,
    ) -> None:
        super().__init__()
        self.name = name
        self.description = description
        self.depends = depends
        self.commands = commands
        self.default_target = default_target
        self.byproducts = byproducts

    def to_string(self) -> str:
        add_to_all_target = "ALL" if self.default_target else ""
        content = [
            CMakeComment(self.description),
            f"add_custom_target({self.name} {add_to_all_target}",
        ]
        content.extend(self._get_commands())
        if self.depends:
            content.append(CMakeDepends(self.depends).to_string())
        if self.byproducts:
            content.append(CMakeByproducts(self.byproducts).to_string())
        content.append(")")
        return "\n".join(str(line) for line in content)

    def _get_commands(self) -> list[str]:
        return [self.tab_prefix + str(command) for command in self.commands]


class CMakeAddSubdirectory(CMakeElement):
    """
    Add_subdirectory(source_dir [binary_dir] [EXCLUDE_FROM_ALL] [SYSTEM]).

    https://cmake.org/cmake/help/latest/command/add_subdirectory.html#add-subdirectory
    """

    def __init__(self, source_dir: CMakePath, build_dir: Optional[CMakePath] = None) -> None:
        super().__init__()
        self.source_dir = source_dir
        self.build_dir = build_dir

    def to_string(self) -> str:
        arguments = [self.source_dir.to_string()]
        if self.build_dir:
            arguments.append(self.build_dir.to_string())
        return "add_subdirectory(" + " ".join(arguments) + ")"


class CMakeListAppend(CMakeElement):
    def __init__(self, variable: str, values: list[str]) -> None:
        super().__init__()
        self.variable = variable
        self.values = values

    def to_string(self) -> str:
        return f"list(APPEND {self.variable} {self._get_values_string()})"

    def _get_values_string(self) -> str:
        return " ".join(self.values)


class CMakeEnableTesting(CMakeElement):
    def __init__(self) -> None:
        super().__init__()

    def to_string(self) -> str:
        return "enable_testing()"



from pathlib import Path
from typing import Optional

from py_app_dev.core.logging import logger


class CMakeRunner:
    executable = "cmake"

    def __init__(self, project_dir: Path, build_dir: Path) -> None:
        self.logger = logger.bind()
        self.project_dir = project_dir
        self.build_dir = build_dir

    def get_configure_command(self, toolchain_file: Optional[str], variant_name: Optional[str], platform_name: Optional[str], build_type: Optional[str] = None) -> list[str | Path]:
        cmake_args = [
            "-S",
            self.project_dir.absolute().as_posix(),
            "-B",
            self.build_dir.absolute().as_posix(),
            "-G",
            "Ninja",
        ]
        if toolchain_file:
            cmake_args.append(f"-DCMAKE_TOOLCHAIN_FILE={toolchain_file}")
        if variant_name:
            cmake_args.append(f"-DVARIANT={variant_name}")
        if platform_name:
            cmake_args.append(f"-DPLATFORM={platform_name}")
        if build_type:
            cmake_args.append(f"-DCMAKE_BUILD_TYPE={build_type}")
        return [self.executable, *cmake_args]

    def get_build_command(self, target: str = "all") -> list[str | Path]:
        return [
            self.executable,
            "--build",
            self.build_dir.absolute().as_posix(),
            "--target",
            target,
            "--",
        ]



import sys
from pathlib import Path
from typing import Optional

import typer
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger, setup_logger, time_it

from yanga import __version__

from .commands.run import RunCommand, RunCommandConfig
from .gui import YangaGui
from .kickstart.create import KickstartProject
from .yide import IDEProjectGenerator
from .yview import KConfigView

package_name = "yanga"

app = typer.Typer(name=package_name, help="YANGA command line interface.", no_args_is_help=True, add_completion=False)


def make_absolute_path(path_value: str | Path) -> Path:
    return Path(path_value).absolute()


project_dir_option = typer.Option(Path.cwd(), callback=make_absolute_path, help="The project directory")


@app.callback(invoke_without_command=True)
def version(
    version: bool = typer.Option(None, "--version", "-v", is_eager=True, help="Show version and exit."),
) -> None:
    if version:
        typer.echo(f"{package_name} {__version__}")
        raise typer.Exit()


@app.command(help="Initialize a new YANGA project in the specified directory.")
@time_it("init")
def init(
    project_dir: Path = project_dir_option,
    force: bool = typer.Option(False, help="Force the initialization of the project even if the directory is not empty."),
    sources: bool = typer.Option(True, help="Copy the example SPL sources."),
) -> None:
    KickstartProject(project_dir, force, sources).run()


@app.command(help="Run the build pipeline for a specific variant and component.")
@time_it("run")
def run(
    project_dir: Path = project_dir_option,
    platform: Optional[str] = typer.Option(
        None,
        help="Platform for which to build (see the available platforms in the configuration).",
    ),
    variant: Optional[str] = typer.Option(
        None,
        help="SPL variant name. If none is provided, it will prompt to select one.",
    ),
    component: Optional[str] = typer.Option(
        None,
        help="Restrict the scope to one specific component.",
    ),
    target: Optional[str] = typer.Option(
        None,
        help="Define a specific target to execute.",
    ),
    build_type: Optional[str] = typer.Option(
        None,
        help="Build type to use (e.g., 'Debug', 'Release').",
    ),
    step: Optional[str] = typer.Option(
        None,
        help="Name of the step to run (as written in the pipeline config).",
    ),
    single: bool = typer.Option(
        False,
        help="If provided, only the provided step will run, without running all previous steps in the pipeline.",
    ),
    print: bool = typer.Option(
        False,
        help="Print the pipeline steps.",
    ),
    force_run: bool = typer.Option(
        False,
        help="Force the execution of a step even if it is not dirty.",
    ),
    not_interactive: bool = typer.Option(
        False,
        help="Run in non-interactive mode (no prompts).",
    ),
) -> None:
    RunCommand().do_run(
        RunCommandConfig(
            project_dir,
            platform,
            variant,
            component,
            target,
            build_type,
            step,
            single,
            print,
            force_run,
            not_interactive,
        )
    )


@app.command(help="Launch the YANGA GUI to build variants and components.")
@time_it("gui")
def gui(
    project_dir: Path = project_dir_option,
) -> None:
    YangaGui(project_dir).run()


@app.command(help="View the variants feature configurations.")
@time_it("view")
def view(
    project_dir: Path = project_dir_option,
) -> None:
    KConfigView(project_dir).run()


@app.command(help="Generate the VS Code project files.")
@time_it("ide")
def ide(
    project_dir: Path = project_dir_option,
) -> None:
    IDEProjectGenerator(project_dir).run()


def main() -> int:
    try:
        setup_logger()
        app()
        return 0
    except UserNotificationException as e:
        logger.error(f"{e}")
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""Command line utility to generate an HTML report from a CppCheck XML report."""

import xml.etree.ElementTree as ET
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from clanguru.doc_generator import CodeContent, DocStructure, MarkdownFormatter, Section, TextContent
from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger

from .base import create_config


@dataclass
class Location(DataClassDictMixin):
    """Represents a location in source code."""

    file: Path
    line: int
    column: int = 0
    info: str = ""


@dataclass
class CppCheckError(DataClassDictMixin):
    """Represents a single CppCheck error/issue."""

    id: str
    severity: str
    msg: str
    verbose: str = ""
    cwe: Optional[str] = None
    file0: Optional[Path] = None
    locations: list[Location] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)


@dataclass
class CppCheckResults(DataClassDictMixin):
    """Representation of the CppCheck XML data."""

    version: str = "2"
    cppcheck_version: str = ""
    errors: list[CppCheckError] = field(default_factory=list)

    def get_errors_by_severity(self) -> dict[str, list[CppCheckError]]:
        """Group errors by severity level."""
        result: dict[str, list[CppCheckError]] = {}
        for error in self.errors:
            if error.severity not in result:
                result[error.severity] = []
            result[error.severity].append(error)
        return result

    def get_errors_by_file(self) -> dict[Path, list[CppCheckError]]:
        """Group errors by source file."""
        result: dict[Path, list[CppCheckError]] = {}
        for error in self.errors:
            # Use file0 if available, otherwise use first location's file
            file_name = error.file0
            if not file_name and error.locations:
                file_name = error.locations[0].file

            if file_name:
                if file_name not in result:
                    result[file_name] = []
                result[file_name].append(error)
        return result

    def get_severity_counts(self) -> dict[str, int]:
        """Get count of errors per severity level."""
        counts: dict[str, int] = {}
        for error in self.errors:
            counts[error.severity] = counts.get(error.severity, 0) + 1
        return counts


def create_doc_structure(cppcheck_data: CppCheckResults, component_name: str, context_lines: int = 3, project_dir: Optional[Path] = None) -> DocStructure:
    """Create a DocStructure from CppCheck results for documentation generation."""
    doc = DocStructure(component_name)

    # Add statistics section
    statistics_section = Section("Statistics")
    total_issues = len(cppcheck_data.errors)
    severity_counts = cppcheck_data.get_severity_counts()

    stats_text = f"Total issues found: {total_issues}\n\n"
    if severity_counts:
        stats_text += "Issues by severity:\n"
        for severity, count in sorted(severity_counts.items()):
            stats_text += f"- {severity}: {count}\n"
    else:
        stats_text += "No issues found."

    statistics_section.add_content(TextContent(stats_text.strip()))
    doc.add_section(statistics_section)

    # Group errors by file and create sections
    errors_by_file = cppcheck_data.get_errors_by_file()

    for file_path, file_errors in sorted(errors_by_file.items()):
        file_section = _create_file_section(file_path, file_errors, context_lines, project_dir)
        doc.add_section(file_section)

    return doc


def _create_file_section(file_path: Path, file_errors: list[CppCheckError], context_lines: int, project_dir: Optional[Path] = None) -> Section:
    """Create a documentation section for a single file with its errors."""
    file_section = Section(file_path.name)

    # Add location information as the first content
    display_path = file_path
    if project_dir is not None:
        try:
            display_path = file_path.relative_to(project_dir)
        except ValueError:
            # If file_path is not relative to project_dir, use absolute path
            display_path = file_path

    location_text = f"Location: {display_path}"
    file_section.add_content(TextContent(location_text))

    for error in file_errors:
        # Create section for each issue
        issue_title = f"{error.severity.title()}: {error.id}"
        issue_section = Section(issue_title)

        # Add issue description
        description = error.verbose or error.msg
        if error.cwe:
            description += f"\n\nCWE: {error.cwe}"

        issue_section.add_content(TextContent(description))

        # Add code context if we have location information
        primary_location = None
        if error.locations:
            primary_location = error.locations[0]
        elif error.file0:
            # If we only have file0, create a basic location
            primary_location = Location(file=error.file0, line=1)

        if primary_location:
            code_context = _extract_code_context(primary_location, context_lines)
            if code_context:
                # If extraction failed (returns error comment) keep previous formatting via CodeContent
                if code_context.startswith("// Could not read file"):
                    issue_section.add_content(CodeContent(code_context, "c"))
                else:
                    # Provide full MyST directive block directly so use TextContent
                    issue_section.add_content(TextContent(code_context))

        file_section.add_subsection(issue_section)

    return file_section


def _extract_code_context(location: Location, context_lines: int) -> str:
    """
    Extract code context around the specified location and format it as a MyST code-block directive.

    Returns a full MyST fenced directive block, e.g.::

        ```{code-block}
        :linenos:
        :lineno-start: 42
        :emphasize-lines: 3
        <code>
        ```

    If the file can't be read, returns an error comment string starting with
    ``// Could not read file`` so callers may decide on fallback formatting.
    """
    try:
        with open(location.file, encoding="utf-8") as file:
            lines = file.readlines()

        target_line_index = location.line - 1  # 0-based
        if target_line_index < 0 or target_line_index >= len(lines):
            return ""

        # Determine slice
        start_index = max(0, target_line_index - context_lines)
        end_index = min(len(lines), target_line_index + context_lines + 1)

        snippet = [lines[i].rstrip("\n") for i in range(start_index, end_index)]

        # Relative line to emphasize inside snippet (1-based within snippet)
        emphasize_relative = (target_line_index - start_index) + 1
        lineno_start = start_index + 1  # original file line number for first snippet line

        # Build MyST code-block directive
        directive_header = [
            "```{code-block}",
            ":linenos:",
            f":lineno-start: {lineno_start}",
            f":emphasize-lines: {emphasize_relative}",
        ]
        directive_body = "\n".join(snippet)
        myst_block = "\n".join(directive_header) + "\n" + directive_body + "\n```"
        return myst_block
    except (FileNotFoundError, OSError, UnicodeDecodeError):
        return f"// Could not read file: {location.file} at line {location.line}"


@dataclass
class CommandArgs(DataClassDictMixin):
    input_file: Path = field(
        metadata={"help": "CppCheck XML report."},
    )
    output_file: Path = field(
        metadata={"help": "Html report."},
    )
    project_dir: Optional[Path] = field(
        default=None,
        metadata={"help": "Project directory to make file paths relative to. If not provided, absolute paths are used."},
    )


class CppCheckReportCommand(Command):
    def __init__(self) -> None:
        super().__init__("cppcheck_report", "Create cppcheck report from the xml results.")
        self.logger = logger.bind()

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, CommandArgs)

    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = create_config(CommandArgs, args)
        self.logger.debug(f"Generating HTML report {config.output_file} from {config.input_file}")

        # Load and parse the XML data
        cppcheck_data = self.load_xml_data(config.input_file)

        # Generate the document structure from the cppcheck results
        doc_structure = create_doc_structure(cppcheck_data, "CppCheck Report", project_dir=config.project_dir)

        # Generate the Markdown report
        config.output_file.write_text(MarkdownFormatter().format(doc_structure))

        return 0

    def load_xml_data(self, xml_file: Path) -> CppCheckResults:
        """Parse CppCheck XML output into structured data."""
        self.logger.debug(f"Parsing XML file: {xml_file}")

        try:
            tree = ET.parse(xml_file)  # noqa: S314
            root = tree.getroot()
        except ET.ParseError as e:
            self.logger.error(f"Failed to parse XML file {xml_file}: {e}")
            raise
        except FileNotFoundError:
            self.logger.error(f"XML file not found: {xml_file}")
            raise

        # Extract version information
        results_version = root.attrib.get("version", "2")
        cppcheck_version = ""

        cppcheck_elem = root.find("cppcheck")
        if cppcheck_elem is not None:
            cppcheck_version = cppcheck_elem.attrib.get("version", "")

        # Parse all errors
        errors: list[CppCheckError] = []
        errors_elem = root.find("errors")
        if errors_elem is not None:
            for error_elem in errors_elem.findall("error"):
                error = self._parse_error_element(error_elem)
                errors.append(error)

        return CppCheckResults(version=results_version, cppcheck_version=cppcheck_version, errors=errors)

    def _parse_error_element(self, error_elem: ET.Element) -> CppCheckError:
        """Parse a single error element into a CppCheckError object."""
        # Extract attributes
        error_id = error_elem.attrib.get("id", "")
        severity = error_elem.attrib.get("severity", "")
        msg = error_elem.attrib.get("msg", "")
        verbose = error_elem.attrib.get("verbose", msg)  # fallback to msg if verbose not present
        cwe = error_elem.attrib.get("cwe")
        file0_elem = error_elem.attrib.get("file0")
        file0 = Path(file0_elem) if file0_elem else None

        # Parse locations
        locations: list[Location] = []
        for loc_elem in error_elem.findall("location"):
            location = Location(
                file=Path(loc_elem.attrib.get("file", "")),
                line=int(loc_elem.attrib.get("line", 0)),
                column=int(loc_elem.attrib.get("column", 0)),
                info=loc_elem.attrib.get("info", ""),
            )
            locations.append(location)

        # Parse symbols
        symbols: list[str] = []
        for symbol_elem in error_elem.findall("symbol"):
            if symbol_elem.text:
                symbols.append(symbol_elem.text)

        return CppCheckError(id=error_id, severity=severity, msg=msg, verbose=verbose, cwe=cwe, file0=file0, locations=locations, symbols=symbols)



import io
import json
import traceback
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

from mashumaro.config import BaseConfig
from mashumaro.mixins.json import DataClassJSONMixin
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger
from py_app_dev.core.scoop_wrapper import ScoopFileElement, ScoopWrapper
from pypeline.domain.pipeline import PipelineStep

from yanga.domain.config import ScoopManifest
from yanga.domain.execution_context import ExecutionContext


@dataclass
class ScoopInstallExecutionInfo(DataClassJSONMixin):
    install_dirs: list[Path] = field(default_factory=list)
    env_vars: dict[str, Any] = field(default_factory=dict)

    class Config(BaseConfig):
        # Make sure to omit None values during serialization
        omit_none = True

    @classmethod
    def from_json_file(cls, file_path: Path) -> "ScoopInstallExecutionInfo":
        try:
            result = cls.from_dict(json.loads(file_path.read_text()))
        except Exception as e:
            output = io.StringIO()
            traceback.print_exc(file=output)
            raise UserNotificationException(output.getvalue()) from e
        return result

    def to_json_string(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def to_json_file(self, file_path: Path) -> None:
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(self.to_json_string())


def create_scoop_wrapper() -> ScoopWrapper:
    return ScoopWrapper()


class ScoopInstall(PipelineStep[ExecutionContext]):
    def __init__(self, execution_context: ExecutionContext, group_name: str, config: Optional[dict[str, Any]] = None) -> None:
        super().__init__(execution_context, group_name, config)
        self.logger = logger.bind()
        self.artifacts_locator = execution_context.create_artifacts_locator()
        self._collected_dependencies: Optional[ScoopManifest] = None
        self.execution_info = ScoopInstallExecutionInfo()
        # One needs to keep track of the installed apps to get the required paths
        # even if the step does not need to run.
        self.execution_info_file = self.output_dir.joinpath("scoop_install_exec_info.json")

    @property
    def output_dir(self) -> Path:
        if self.execution_context.variant:
            return self.execution_context.create_artifacts_locator().variant_build_dir
        else:
            return self.artifacts_locator.build_dir

    def get_name(self) -> str:
        return self.__class__.__name__

    @property
    def install_dirs(self) -> list[Path]:
        return self.execution_info.install_dirs

    @property
    def scoop_manifest_file(self) -> Path:
        """Generated scoopfile.json file path in variant-specific directory."""
        return self.artifacts_locator.variant_build_dir.joinpath("scoopfile.json")

    @property
    def global_scoop_manifest_file(self) -> Path:
        """Global scoopfile.json file path (if exists)."""
        return self.project_root_dir.joinpath("scoopfile.json")

    def _collect_dependencies(self) -> ScoopManifest:
        """Collect Scoop dependencies from global, platform, and variant configurations."""
        collected_manifest = ScoopManifest()

        # Start with global scoopfile.json if it exists
        if self.global_scoop_manifest_file.exists():
            try:
                global_manifest_file = ScoopManifest.from_file(self.global_scoop_manifest_file)

                # Merge buckets
                self._merge_buckets(collected_manifest, global_manifest_file.buckets)

                # Merge apps
                self._merge_apps(collected_manifest, global_manifest_file.apps)

            except Exception as e:
                self.logger.warning(f"Failed to parse global scoopfile.json: {e}")

        # Add platform dependencies
        if self.execution_context.platform and self.execution_context.platform.scoop_manifest:
            platform_manifest = self.execution_context.platform.scoop_manifest

            # Merge buckets
            self._merge_buckets(collected_manifest, platform_manifest.buckets)

            # Merge apps
            self._merge_apps(collected_manifest, platform_manifest.apps)

        # Add variant dependencies
        if self.execution_context.variant and self.execution_context.variant.scoop_manifest:
            variant_manifest = self.execution_context.variant.scoop_manifest

            # Merge buckets
            self._merge_buckets(collected_manifest, variant_manifest.buckets)

            # Merge apps
            self._merge_apps(collected_manifest, variant_manifest.apps)

        return collected_manifest

    def _merge_buckets(self, target_manifest: ScoopManifest, source_buckets: list[ScoopFileElement]) -> None:
        """Merge buckets, handling conflicts when same name has different sources."""
        for bucket in source_buckets:
            # Check if a bucket with this name already exists
            existing_bucket = next((b for b in target_manifest.buckets if b.name == bucket.name), None)

            if existing_bucket is None:
                # No conflict, add the bucket
                target_manifest.buckets.append(bucket)
            elif existing_bucket.source != bucket.source:
                # Conflict: same name, different source
                self.logger.warning(
                    f"Bucket '{bucket.name}' defined multiple times with different sources:\n"
                    f"  Existing: {existing_bucket.source}\n"
                    f"  New: {bucket.source}\n"
                    f"  Keeping existing definition."
                )
                # Keep the first definition (existing_bucket), ignore the new one

    def _merge_apps(self, target_manifest: ScoopManifest, source_apps: list[ScoopFileElement]) -> None:
        """Merge apps, avoiding duplicates."""
        for app in source_apps:
            if app not in target_manifest.apps:
                target_manifest.apps.append(app)

    def _generate_scoop_manifest(self, manifest: ScoopManifest) -> None:
        """Generate scoopfile.json file from collected dependencies."""
        if not manifest.buckets and not manifest.apps:
            self.logger.info("No Scoop dependencies found. Skipping scoopfile.json generation.")
            return

        # Ensure build directory exists
        self.scoop_manifest_file.parent.mkdir(parents=True, exist_ok=True)
        self.scoop_manifest_file.write_text(manifest.to_json_string())

        self.logger.info(f"Generated scoopfile.json with {len(manifest.buckets)} buckets and {len(manifest.apps)} apps")

    def run(self) -> int:
        self.logger.debug(f"Run {self.get_name()} step. Output dir: {self.output_dir}")

        try:
            # Collect dependencies from all sources
            self._collected_dependencies = self._collect_dependencies()

            # Generate the combined scoopfile.json file
            self._generate_scoop_manifest(self._collected_dependencies)

            # Only proceed with scoop commands if we have dependencies
            if not self._collected_dependencies.apps:
                self.logger.info("No Scoop dependencies to install.")
                return 0

            # Install scoop dependencies using the generated manifest file
            installed_apps = create_scoop_wrapper().install(self.scoop_manifest_file)

            self.logger.debug("Installed apps:")
            for app in installed_apps:
                self.logger.debug(f" - {app.name} ({app.version})")
                self.execution_info.install_dirs.extend(app.get_all_required_paths())
                # Collect environment variables from each app
                self.execution_info.env_vars.update(app.env_vars)

            # Save execution info for dependency tracking
            self.execution_info.to_json_file(self.execution_info_file)

        except Exception as e:
            raise UserNotificationException(f"Failed to install scoop dependencies: {e}") from e

        return 0

    def get_inputs(self) -> list[Path]:
        inputs = []

        # Add global scoopfile.json if it exists
        if self.global_scoop_manifest_file.exists():
            inputs.append(self.global_scoop_manifest_file)

        # Add user config files (platform and variant configs are in there)
        inputs.extend(self.execution_context.user_config_files)

        return inputs

    def get_outputs(self) -> list[Path]:
        outputs = [self.scoop_manifest_file, self.execution_info_file]

        # Add tracked install directories if available (after run() has been called)
        if self.execution_info.install_dirs:
            outputs.extend(self.execution_info.install_dirs)

        return outputs

    def update_execution_context(self) -> None:
        if self.execution_info_file.exists():
            execution_info = ScoopInstallExecutionInfo.from_json_file(self.execution_info_file)
            # Make the list unique and keep the order
            unique_paths = list(dict.fromkeys(execution_info.install_dirs))
            # Update the install directories for the subsequent steps
            self.execution_context.add_install_dirs(unique_paths)
            if execution_info.env_vars:
                self.execution_context.add_env_vars(execution_info.env_vars)



import sys
from pathlib import Path
from typing import Optional

from py_app_dev.core.exceptions import UserNotificationException


class ProjectArtifactsLocator:
    """Provides paths to project artifacts."""

    def __init__(
        self,
        project_root_dir: Path,
        variant_name: Optional[str],
        platform_name: Optional[str],
        build_type: Optional[str],
        create_yanga_build_dir: bool = True,
    ) -> None:
        self.project_root_dir = project_root_dir
        yanga_out_dir = project_root_dir / ".yanga"
        self.build_dir = yanga_out_dir / "build" if create_yanga_build_dir else project_root_dir / "build"
        self.variants_dir = project_root_dir / "variants"
        self.platforms_dir = project_root_dir / "platforms"
        self.variant_build_dir: Path = self.determine_variant_build_dir(variant_name, platform_name, build_type, self.build_dir)
        self.variant_dir: Optional[Path] = self.variants_dir / variant_name if variant_name else None
        # We do not need to create an extra directory if the .yanga build dir is used
        self.external_dependencies_dir = yanga_out_dir / "ext" if create_yanga_build_dir else self.build_dir / "ext"
        scripts_dir = "Scripts" if sys.platform.startswith("win32") else "bin"
        self.venv_scripts_dir = self.project_root_dir.joinpath(".venv").joinpath(scripts_dir)

    def locate_artifact(self, artifact: str, first_search_paths: list[Optional[Path]]) -> Path:
        search_paths: list[Optional[Path]] = []
        for path in first_search_paths:
            if path:
                search_paths.append(path.parent if path.is_file() else path)
        search_paths.extend(
            [
                self.variant_dir,
                self.project_root_dir,
                self.platforms_dir,
            ]
        )
        for dir in search_paths:
            if dir and (artifact_path := Path(dir).joinpath(artifact)).exists():
                return artifact_path
        else:
            raise UserNotificationException(f"Artifact '{artifact}' not found in the project. Searched paths: {', '.join(str(p) for p in search_paths if p is not None)}")

    @staticmethod
    def determine_variant_build_dir(variant_name: Optional[str], platform_name: Optional[str], build_type: Optional[str], build_dir: Path) -> Path:
        # build up the path in order: variant, platform, build_type
        parts = []
        if variant_name:
            parts.append(variant_name)
        if platform_name:
            parts.append(platform_name)
        if build_type:
            parts.append(build_type)
        return build_dir.joinpath(*parts) if parts else build_dir
